/* KeyChoice.java
 *
 * created: Mon Sep  6 1999
 *
 * This file is part of Artemis
 * 
 * Copyright (C) 1999  Genome Research Limited
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 * $Header: /nfs/disk222/yeastpub/Repository/powmap/diana/components/KeyChoice.java,v 1.2 1999/09/09 11:23:13 kmr Exp $
 */

package diana.components;

import diana.*;

import uk.ac.sanger.pathogens.StringVector;
import uk.ac.sanger.pathogens.embl.Key;
import uk.ac.sanger.pathogens.embl.InvalidKeyException;
import uk.ac.sanger.pathogens.embl.EntryInformation;

import java.awt.*;
import java.awt.event.*;

/**
 *  This component is a Choice component that shows the possible feature keys.
 *
 *  @author Kim Rutherford
 *  @version $Id: KeyChoice.java,v 1.2 1999/09/09 11:23:13 kmr Exp $
 **/

public class KeyChoice extends Panel {
  /**
   *  Create a new KeyChoice component with CDS as the default key.
   **/
  public KeyChoice () {
    this (Key.CDS);
  }

  /**
   *  Create a new KeyChoice component with the given key as the default.
   **/
  public KeyChoice (final Key default_key) {
//      final Font font = Options.getOptions ().getFont ();

//      key_chooser.setFont (font);

    this.default_key = default_key;

    updateChoice ();

    common_keys_checkbox.addItemListener (new ItemListener () {
      public void itemStateChanged(ItemEvent e) {
        updateChoice ();
      }
    });

    add (key_chooser);
    add (common_keys_checkbox);
  }

  /**
   *  Return the currently selected key.
   **/
  public Key getSelectedItem () {
    final Key key;

    try {
      key = new Key (key_chooser.getSelectedItem ());
    } catch (InvalidKeyException e) {
      throw new Error ("internal error - " +
                       "InvalidKeyException in KeyChoice.getSelectedItem ()");
    }

    return key;
  }

  /**
   *  Set the selected Key.
   **/
  public void setKey (final Key new_key) {
    final int key_index = keyIndex (new_key);

    if (key_index == -1) {
      // add the key
      key_chooser.add (new_key.toString ());
      key_chooser.select (new_key.toString ());
    } else {
      key_chooser.select (key_index);
    }
  }

  /**
   *  Adds the specified item listener to receive item events from the Choice
   *  component of this KeyChoice.
   *  @param l The item listener.
   **/
  public void addItemListener(ItemListener l) {
    key_chooser.addItemListener (l);
  }


  /**
   *  Removes the specified item listener so that it no longer receives item
   *  events from the Choice component of this KeyChoice.
   *  @param l The item listener.
   **/
  public void removeItemListener(ItemListener l) {
    key_chooser.removeItemListener (l);
  }

  /**
   *  Update the key_chooser.
   **/
  private void updateChoice () {
    while (key_chooser.getItemCount () > 0) {
      key_chooser.remove (key_chooser.getItemCount () - 1);
    }

    final Font font = Options.getOptions ().getFont ();

    key_chooser.setFont (font);

    final StringVector key_strings;

    if (common_keys_checkbox.getState ()) {
      key_strings = getSortedCommonKeys ();
    } else {
      key_strings = getSortedUncommonKeys ();
    }

    for (int i = 0 ; i < key_strings.size () ; ++i) {
      key_chooser.add (key_strings.elementAt (i));
    }

    if (keyIndex (default_key) != -1) {
      setKey (default_key);
    }
  }

  /**
   *  Return a alphanumerically sorted vector containing the String
   *  representations of the common keys (those listed in the common_keys
   *  property of the options file).  If there is no common_keys option then
   *  all the legal keys are returned.
   **/
  private static StringVector getSortedCommonKeys () {
    final StringVector return_vector =
      Options.getOptions ().getOptionValues ("common_keys");

    if (return_vector == null) {
      return EntryInformation.getSortedPossibleKeys ();
    }

    return_vector.sort ();
    return return_vector;
  }

  /**
   *  Return a alphanumerically sorted vector containing the String
   *  representations of the uncommon keys (those not listed in the
   *  common_keys property of the options file).  If there is no common_keys
   *  option then no keys are returned.
   **/
  private static StringVector getSortedUncommonKeys () {
    final StringVector common_keys =
      Options.getOptions ().getOptionValues ("common_keys");

    final StringVector possible_keys =
      EntryInformation.getSortedPossibleKeys ();

    if (common_keys == null) {
      return new StringVector ();
    }

    final StringVector return_vector = new StringVector ();

    for (int i = 0 ; i < possible_keys.size () ; ++i) {
      if (!common_keys.contains (possible_keys.elementAt (i))) {
        return_vector.add (possible_keys.elementAt (i));
      }
    }
    
    return_vector.sort ();
    return return_vector;
  }

  /**
   *  Return the index in the key_chooser component of the given Key.
   **/
  private int keyIndex (final Key key) {
    for (int i = 0 ; i < key_chooser.getItemCount () ; ++i) {
      if (key.toString ().equals (key_chooser.getItem (i))) {
        return i;
      }
    }
    return -1;
  }

  /**
   *  The Key that was passed to the constructor.
   **/
  private Key default_key = null;

  /**
   *  The Choice component that will show the feature keys.
   **/
  private Choice key_chooser = new Choice ();

  /**
   *  This toggle sets whether to show common keys or uncommon keys.
   **/
  private Checkbox common_keys_checkbox =
    new Checkbox ("Common Keys", true);
}
