/* GFFDocumentEntry.java
 *
 * created: Tue Sep 14 1999
 *
 * This file is part of Artemis
 *
 * Copyright (C) 1999  Genome Research Limited
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 * $Header: /nfs/disk222/yeastpub/Repository/powmap/uk/ac/sanger/pathogens/embl/GFFDocumentEntry.java,v 1.1 1999/09/15 15:31:51 kmr Exp $
 */

package uk.ac.sanger.pathogens.embl;

import uk.ac.sanger.pathogens.*;

import java.io.*;

/**
 *  A DocumentEntry that can read an GFF entry from a Document.
 *
 *  @author Kim Rutherford
 *  @version $Id: GFFDocumentEntry.java,v 1.1 1999/09/15 15:31:51 kmr Exp $
 **/

public class GFFDocumentEntry extends EMBLObject
    implements DocumentEntry {
  /**
   *  Create a new GFFDocumentEntry object associated with the given
   *  Document.
   *  @param document This is the file that we will read from.  This is also
   *    used for saving the entry back to the file it came from and to give
   *    the new object a name.
   *  @exception IOException thrown if there is a problem reading the entry -
   *    most likely ReadFormatException.
   *  @exception InvalidRelationException Thrown if a if a Feature in this
   *    entry has a illegal key/qualifier combination.
   **/
  GFFDocumentEntry (Document document)
      throws IOException, InvalidRelationException {
    this.document = document;

    if (! document.readable ()) {
      // this is a new empty GFFDocumentEntry
      return;
    }

    final Reader in_file = document.getReader ();

    LinePushBackReader pushback_reader = new LinePushBackReader (in_file);

    String line = null;

    feature_table = new GFFFeatureTable (this, pushback_reader);

    System.gc();

    in_file.close ();
  }

  /**
   *  Create a new GFFDocumentEntry that will be a copy of the given
   *  Entry and has no Document associated with it.  The new
   *  GFFDocumentEntry cannot be saved to a file with save () unless save
   *  (Document) has been called first.  The save (Document) method will
   *  assign a Document..
   **/
  public GFFDocumentEntry (final Entry new_entry) {
    throw new Error ("internal error - unimplemented");
  }

  /**
   *  Create a new empty GFFDocumentEntry object that has no Document
   *  associated with it.  The new GFFDocumentEntry cannot be saved to a
   *  file with save () unless save (Document) has been called first.  The
   *  save (Document) method will assign a Document.
   **/
  public GFFDocumentEntry () {
    this.document = null;
  }

  /**
   *  Returns true if and only if this entry is read only.  For now this
   *  always returns true - GFFDocumentEntry objects can't be changed.
   **/
  public boolean isReadOnly () {
    return true;
  }

  /**
   *  Return the text of the EMBL header of this Entry or null if there is no
   *  header.
   **/
  public String getHeaderText () {
    return null;
  }

  /**
   *  Set the header of this Entry to be the given text.
   *  @return true if and only if the header was successfully set.  Not all
   *    Entry objects can change their header, so it is up to the calling
   *    function to check the return value.
   *  @exception IOException thrown if there is a problem reading the header
   *    from the String - most likely ReadFormatException.
   **/
  public boolean setHeaderText (final String new_header)
      throws IOException {
    return false;
  }

  /**
   *  Write this Entry to the Document it came from.  This method uses the
   *  current Document reference (as given by getDocument ()).  This method
   *  will throw a NullPointerException if the is no current Document
   *  ie. if getDocument () return null.  Use save (Document) to save and set
   *  the current Document.
   *  @exception IOException thrown if there is a problem saving the entry.
   **/
  public void save ()
      throws IOException {

//    System.out.println ("in GFFDocumentEntry.save ()");sequence.seq
    save (true);
  }

  /**
   *  Write this Entry to the Document it came from.  This method uses the
   *  current Document reference (as given by getDocument ()).  This method
   *  will throw a NullPointerException if the is no current Document
   *  ie. if getDocument () return null.  Use save (Document) to save and set
   *  the current Document.
   *  @param include_diana_extensions If true then the any diana additions to
   *    the embl file format will be included in the output, otherwise they
   *    will be removed.
   *  @exception IOException thrown if there is a problem saving the entry.
   **/
  public void save (final boolean include_diana_extensions)
      throws IOException {

    final Writer out_file = getDocument ().getWriter ();
    writeToStream (out_file);
    out_file.close ();
  }

  /**
   *  Write this Entry to the given Document.
   *  @param document This is the file that we will write to.
   *  @param include_diana_extensions If true then the any diana additions to
   *    the embl file format will be included in the output, otherwise they
   *    will be removed.
   *  @exception IOException thrown if there is a problem saving the entry.
   **/
  public void save (final Document document,
                    final boolean include_diana_extensions)
      throws IOException {

    final Writer out_file = document.getWriter ();
    writeToStream (out_file);
    out_file.close ();
  }

  /**
   *  Write this Entry to the given stream.  The output will be in EMBL
   *  format (but see the diana_extensions argument).
   *  @param writer The stream to write to.
   *  @exception IOException thrown if there is a problem writing the entry.
   **/
  protected void writeToStream (final Writer writer)
      throws IOException {
    final FeatureEnumeration feature_enum = features ();

    while (feature_enum.hasMoreFeatures ()) {
      final GFFStreamFeature new_feature =
        (GFFStreamFeature) feature_enum.nextFeature ();

      new_feature.writeToStream (writer, false);
    }
  }

  /**
   *  Return the File reference that was passed to the constructor or null if
   *  none was passed.
   **/
  public Document getDocument () {
    return document;
  }

  /**
   *  Return a count of the number of Feature obejcts in this Entry.
   **/
  public int getFeatureCount () {
    return feature_table.getFeatureCount ();
  }

  /**
   *  Add the given Feature to this Entry.
   **/
  public Feature add (Feature new_feature) {
    if (new_feature.getEntry () != null) {
      throw new Error ("internal error - a feature must have one owner");
    }

    if (new_feature instanceof GFFStreamFeature) {
      final GFFStreamFeature gff_feature = (GFFStreamFeature) new_feature;
      getFeatureTable ().add (gff_feature);
      gff_feature.setDocumentEntry (this);
      return gff_feature;
    } else {
      throw new Error ("internal error - unimplemented");
    }
  }

  /**
   *  Remove the given Feature from this Entry.
   *  @return true if and only if the Feature was in this Entry.
   **/
  public boolean remove (Feature feature)
      throws ReadOnlyException {
    throw new ReadOnlyException ();
  }

  /**
   *  Return the ith Feature from this Entry.  This Features are returned in a
   *  consistent order, sorted by the first base of each Feature.
   **/
  public Feature getFeatureAtIndex (int i) {
    return feature_table.getFeatureAtIndex (i);
  }

  /**
   *  Return the index of the given Feature.  This does the reverse of
   *  getFeatureAtIndex ().
   **/
  public int indexOf (final Feature feature) {
    return feature_table.indexOf (feature);
  }

  /**
   *  Returns true if and only if this Entry contains the given feature.
   **/
  public boolean contains (final Feature feature) {
    return feature_table.contains (feature);
  }

  /**
   *  Create a new Feature object of an appropriate type in this Entry.
   *  @param key The new feature key
   *  @param location The Location object for the new feature
   *  @param qualifiers The qualifiers for the new feature (can be null if
   *    there are no qualifiers).
   *  @exception InvalidRelationException Thrown if this Feature cannot contain
   *    the given Qualifier.
   **/
  public Feature createFeature (Key key,
                                Location location,
                                QualifierVector qualifiers)
      throws InvalidRelationException, ReadOnlyException {

    throw new ReadOnlyException ();
  }

  /**
   *  Return a vector containing the references of the Feature objects within
   *  the given range.
   *  @param range Return features that overlap this range - ie the start of
   *    the feature is less than or equal to the end of the range and the end
   *    of the feature is greater than or equal to the start of the range.
   *  @return The non-source key features of this feature table the are within
   *    the given range.  The returned object is a copy - changes will not
   *    effect the FeatureTable object itself.
   **/
  public FeatureVector getFeaturesInRange (Range range) {
    return feature_table.getFeaturesInRange (range);
  }

  /**
   *  Return a vector containing the references of all the Feature objects in
   *  this Entry.
   *  @return The non-source key features of this Entry.  The returned object
   *    is a copy - changes will not effect the Entry object itself.
   **/
  public FeatureVector getAllFeatures () {
    return feature_table.getAllFeatures ();
  }

  /**
   *  Returns an enumeration of the Feature objects in this
   *  GFFDocumentEntry.  The returned Enumeration object will generate
   *  all features in this object in turn. The first item generated is the
   *  item at index 0, then the item at index 1, and so on.
   **/
  public FeatureEnumeration features () {
    return feature_table.features ();
  }

  /**
   *  Return the Sequence object from this entry or null if it does not
   *  contain one.
   *  @return a Sequence object for this Entry.  the returned object is
   *    not a copy - changes to it will change the Entry object itself
   **/
  public Sequence getSequence () {
    return null;
  }

  /**
   *  Return the name of this Entry or null if it has no name.
   **/
  public String getName () {
    if (getDocument () == null || getDocument ().getName () == null) {
      return null;
    } else {
      return getDocument ().getName ();
    }
  }

  /**
   *  Set the name of this Entry - if possible (the return value will let the
   *  caller know).
   *  @return true if and only if the name was successfully set.  Not all
   *    Entry objects can change there name, so it is up to the calling
   *    function to check the return value.
   **/
  public boolean setName (final String name) {
    if (name == null || name.length () == 0) {
      return false;
    }
    document = new FileDocument (new File (name));
    return true;
  }

  /**
   *  Return the FeatureTable object from this entry.
   *  @return a FeatureTable object for this Entry.  the returned object is
   *    not a copy - changes to it will change the Entry object itself
   **/
  FeatureTable getFeatureTable () {
    return feature_table;
  }

  /**
   *  The Document object that was passed to the constructor.  This should be
   *  the document that this GFFDocumentEntry was read from.
   **/
  private Document document = null;

  /**
   *  The FeatureTable that contains the features of this DocumentEntry
   **/
  private FeatureTable feature_table = null;
}
