/* GFFStreamFeature.java
 *
 * created: Tue Sep 14 1999
 *
 * This file is part of Artemis
 *
 * Copyright (C) 1999  Genome Research Limited
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 * $Header: /nfs/disk222/yeastpub/Repository/powmap/uk/ac/sanger/pathogens/embl/GFFStreamFeature.java,v 1.1 1999/09/15 15:32:07 kmr Exp $
 */

package uk.ac.sanger.pathogens.embl;

import uk.ac.sanger.pathogens.*;

import java.io.*;

/**
 *  A StreamFeature that thinks it is a GFF feature.
 *
 *  @author Kim Rutherford
 *  @version $Id: GFFStreamFeature.java,v 1.1 1999/09/15 15:32:07 kmr Exp $
 **/

public class GFFStreamFeature extends EMBLObject
    implements StreamFeature, ComparableFeature {
  /**
   *  Create a new GFFStreamFeature from the given line.  The String should be
   *  in gene finder format.
   **/
  private GFFStreamFeature (final String line)
      throws ReadFormatException, LocationParseException {
    final StringVector line_bits = StringVector.getStrings (line);

    if (line_bits.size () < 7) {
      System.err.println (line_bits.size ());
      throw new ReadFormatException ("invalid GFF line (not enough " +
                                     "fields): " +
                                     joinStringVector (line_bits));
    }

    final String start_base = line_bits.elementAt (3);
    final String end_base = line_bits.elementAt (4);

    final boolean complement_flag;

    if (line_bits.elementAt (6).equals ("+")) {
      complement_flag = false;
    } else {
      if (line_bits.elementAt (6).equals ("-")) {
        complement_flag = true;
      } else {
        throw new ReadFormatException ("invalid GFF line (7th field should " +
                                       "+ or -): " +
                                       joinStringVector (line_bits));
      }
    }

    try {
      final Qualifier note_qualifier =
        new Qualifier ("note", joinStringVector (line_bits));

      qualifiers.setQualifier (note_qualifier);

      final Qualifier label_qualifier =
        new Qualifier ("label", line_bits.elementAt (0));

      qualifiers.setQualifier (label_qualifier);

      final Qualifier seqname_qualifier =
        new Qualifier ("gff_seqname", line_bits.elementAt (0));

      qualifiers.setQualifier (seqname_qualifier);

      final Qualifier source_qualifier =
        new Qualifier ("gff_source", line_bits.elementAt (1));

      qualifiers.setQualifier (source_qualifier);

      final Qualifier gff_feature_qualifier =
        new Qualifier ("gff_feature", line_bits.elementAt (2));

      qualifiers.setQualifier (gff_feature_qualifier);

      final Qualifier score_qualifier =
        new Qualifier ("gff_score", line_bits.elementAt (5));

      qualifiers.setQualifier (score_qualifier);

      if (complement_flag) {
        location = new Location ("complement(" + start_base + ".." +
                                 end_base + ")");
      } else {
        location = new Location (start_base + ".." + end_base);
      }
    } catch (QualifierParseException e) {
      throw new Error ("internal error -unexpected exception");
    } catch (InvalidQualifierException e) {
      throw new Error ("internal error -unexpected exception");
    }

    this.gff_line = line;
  }

  /**
   *  Helper method for the constructor - returns a String that is the
   *  concatenation of the Strings in the given StringVector.  The strings
   *  will be separated by four spaces
   **/
  private String joinStringVector (final StringVector string_vector) {
    final StringBuffer return_buffer = new StringBuffer ();

    for (int i = 0 ; i < string_vector.size () ; ++i) {
      if (i != 0) {
        return_buffer.append ("    ");
      }
      return_buffer.append (string_vector.elementAt (i));
    }

    return return_buffer.toString ();
  }

  /**
   *  Set the value of this object.
   *  @param key The new feature key
   *  @param location The Location object for the new feature
   *  @param qualifiers The qualifiers for the new feature
   *  @exception InvalidRelationException Thrown if this Feature cannot contain
   *    any of the given Qualifier objects.
   **/
  public void set (Key key,
                   Location location,
                   QualifierVector qualifiers)
      throws InvalidRelationException, ReadOnlyException {
    throw new ReadOnlyException ();
  }

  /**
   *  Set the owning Entry of this Feature.  Other objects should call
   *  setDocumentEntry () to change the owner.
   *  @param entry The Entry that now owns this Feature.
   **/
  private void setEntry (GFFDocumentEntry entry) {
    this.entry = entry;
  }


  /**
   *  Set the owning DocumentEntry of this Feature.
   *  @param entry The Entry that now owns this Feature.
   **/
  void setDocumentEntry (DocumentEntry entry) {
    setEntry ((GFFDocumentEntry) entry);
  }

  /**
   *  Set the key field of this object.
   *  @param key The new feature key
   **/
  public void setKey (Key key)
      throws ReadOnlyException {
    throw new ReadOnlyException ();
  }

  /**
   *  Set the location of this object.
   *  @param location The Location object for the new feature
   **/
  public void setLocation (Location location)
      throws ReadOnlyException {
    throw new ReadOnlyException ();
  }

  /**
   *  Set the qualifiers of this object discarding all the current qualifiers.
   *  @param qualifiers The qualifiers for the new feature
   *  @exception InvalidRelationException Throw if this Feature cannot contain
   *    any of the given Qualifier objects.
   **/
  public void setQualifiers (QualifierVector qualifiers)
      throws InvalidRelationException, ReadOnlyException {
    throw new ReadOnlyException ();
  }

  /**
   *  Add the given Qualifier to this Feature.  If this Feature contains a
   *  Qualifier with the same name as the new Qualifier it will be replaced.
   *  @param qualifier The new qualifier to add.
   *  @exception InvalidRelationException Throw if this Feature cannot contain
   *    the given Qualifier.
   **/
  public void setQualifier (Qualifier qualifier)
      throws InvalidRelationException, ReadOnlyException {
    throw new ReadOnlyException ();
  }

  /**
   *  Remove the Qualifier with the given name.  If there is no Qualifier with
   *  that name then return immediately.
   *  @param name The qualifier name to look for.
   **/
  public void removeQualiferByName (final String name)
      throws ReadOnlyException {
    throw new ReadOnlyException ();
  }

  /**
   *  Add the values from the given qualifier to the Qualifier object with the
   *  same name in this Feature or otherwise add a copy of the argument.
   *  @param qualifier This object contians name and values to add.
   *  @param values The values to add to the Qualifier.
   *  @return The Qualifier that was changed or created.
   *  @exception InvalidRelationException Thrown if this Feature cannot contain
   *    the given Qualifier.
   **/
  public Qualifier addQualifierValues (Qualifier qualifier)
      throws InvalidRelationException, ReadOnlyException {
    throw new ReadOnlyException ();
  }

  /**
   *  Return the unique identifier of this StreamFeature.
   **/
  public long getNumericID () {
    return id;
  }

  /**
   *  Return the key of this feature, as passed to the constructor.
   **/
  public Key getKey () {
    return gff_key;
  }

  /**
   *  Return the Location of this Feature, as passed to the constructor.
   **/
  public Location getLocation () {
    return location;
  }

  /**
   *  Return a QualifierVector object containing the qualifiers for this
   *  feature.  This method does not return a copy of the qualifier vector so
   *  changing the vector will change the feature.
   **/
  public QualifierVector getQualifiers () {
    return qualifiers;
  }

  /**
   *  Return the Qualifier in this Feature with the given name or null if
   *  there no such Qualifier.
   **/
  public Qualifier getQualifierByName (final String name) {
    return getQualifiers ().getQualifierByName (name);
  }

  /**
   *  Return the first base of this feature.
   **/
  public int getFirstBase () {
    return getLocation ().getFirstBase ();
  }

  /**
   *  Return the last base of this feature.
   **/
  public int getLastBase () {
    return getLocation ().getLastBase ();
  }

  /**
   *  Return the Entry object that contains this Feature.
   **/
  public Entry getEntry () {
    return entry;
  }

  /**
   *  Return the reference of a new copy of this Feature.  The new Feature
   *  will not be in any Entry, so it should be explicitly added with a call
   *  to Entry.add ().
   **/
  public Feature copy () {
    // XXX
    return null;
  }

  /**
   *  Write this GFFStreamFeature to the given stream.  The output will be in
   *  GFF format (but see the diana_extensions argument).
   *  @param writer The stream to write to.
   *  @param diana_extensions This argument is ignored.
   *  @exception IOException thrown if there is an io problem while writing
   *    the Feature.
   **/
  public void writeToStream (final Writer writer,
                             final boolean diana_extensions)
      throws IOException {

    writer.write (gff_line);
  }

  /**
   *  Read and return a GFFStreamFeature from a stream.  A feature must be the
   *  next thing in the stream.
   *  @param stream the Feature is read from this stream
   *  @exception IOException thrown if there is a problem reading the Feature -
   *    most likely ReadFormatException.
   *  @exception InvalidRelationException Thrown if this Feature cannot contain
   *    the given Qualifier.
   *  @return null if in_stream is at the end of file when the method is
   *    called
   */
  protected static GFFStreamFeature readFromStream (LinePushBackReader stream)
      throws IOException, InvalidRelationException {

    String line = stream.readLine ();

    if (line == null) {
      return null;
    }

    while (line != null && line.startsWith ("#")) {
      line = stream.readLine ();
    }

    if (line == null) {
      return null;
    }

    try {
      final GFFStreamFeature new_feature = new GFFStreamFeature (line);

      return new_feature;
    } catch (LocationParseException exception) {
      // re-throw the exception with the line number added

      final String new_error_string = exception.getMessage ();

      System.err.println (new_error_string);

      throw new ReadFormatException (new_error_string,
                                     stream.getLineNumber ());
    } catch (ReadFormatException exception) {
      // re-throw the exception with the line number added

      final String new_error_string = exception.getMessage ();

      System.err.println (new_error_string);

      throw new ReadFormatException (new_error_string,
                                     stream.getLineNumber ());
    }
  }

  /**
   *  Read the details of a feature from an EMBL stream into the current
   *  object.
   *  @param in_stream the Feature is read from this stream
   *  @exception IOException thrown if there is a problem reading the Feature -
   *    most likely ReadFormatException if the stream does not contain GFF
   *    feature.
   **/
  public void setFromStream (final LinePushBackReader in_stream)
      throws IOException, ReadOnlyException {
    throw new ReadOnlyException ();
  }

  /**
   *  The DocumentEntry object that contains this Feature as passed to the
   *  constructor.
   **/
  private DocumentEntry entry;

  /**
   *  The is the Location object that was passed to the constructor
   **/
  private Location location;

  /**
   *  The is the vector of qualifiers that was passed to the constructor
   **/
  private QualifierVector qualifiers = new QualifierVector ();

  /**
   *  This is incremented each time a constructor is called.
   **/
  private static long id_counter = 0;

  /**
   *  A unique identifier for this feature.
   **/
  private final long id = id_counter++;

  /**
   *  The Key to use for all GFFStreamFeature objects.
   **/
  private static Key gff_key = new Key ("GFF", true);

  /**
   *  This is the line of GFF input that was read to get this GFFStreamFeature.
   **/
  private String gff_line = null;
}
