/* Location.java
 *
 * created: Mon Oct  5 1998
 *
 * This file is part of Artemis
 *
 * Copyright (C) 1998,1999  Genome Research Limited
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 * $Header: /nfs/disk222/yeastpub/Repository/powmap/uk/ac/sanger/pathogens/embl/Location.java,v 1.3 1999/09/28 13:08:29 kmr Exp $
 */

package uk.ac.sanger.pathogens.embl;

import uk.ac.sanger.pathogens.*;
import uk.ac.sanger.pathogens.embl.LocationLexer.*;

import java.util.Vector;

/**
 *  This class encapsulates the location field of a EMBL entry feature.
 *  As well as allowing access to the location string itself, this class has
 *  functions for parsing and manipulating the location.
 *
 *  @author Kim Rutherford
 *  @version $Id: Location.java,v 1.3 1999/09/28 13:08:29 kmr Exp $
 *
 */

public class Location {
  /**
   *  Constructs a new Location object with that represents the value
   *  represented by the string.
   *  @exception LocationParseException Thrown if the location String does not
   *    represent a valid Location.
   **/
  public Location (final String location_string)
      throws LocationParseException {
    parse_tree = getParseTree (location_string).getCanonical ();

    if (parse_tree == null) {
      throw new LocationParseException ("invalid location", location_string);
    }
  }

  /**
   *  Constructs a new Location object with that represents the given Range.
   *  @exception LocationParseException Thrown is the Range start below one.
   **/
  public Location (final Range location_range)
      throws LocationParseException {
    if (location_range.getStart () < 1) {
      throw new LocationParseException ("location out of range",
                                        location_range.toString ());
    } else {
      parse_tree = new LocationParseNode (location_range);
    }
  }

  /**
   *  Returns the value of this Location as a string.  The value returned will
   *  be the same as string passed to the constructor (but will always look
   *  like join(complement(... if there is a complement and a join).
   **/
  public String toString () {
    return getParsedLocation ().toString ();
  }

  /**
   *  Return true if and only if this Location is the same as the given
   *  Location.
   **/
  public boolean equals (final Location other_location) {
    return toString ().equals (other_location.toString ());
  }

  /**
   *  Returns the value of this Location as a string with the complement (if
   *  any) as the top node and the join (if any) below it.  The value returned
   *  will be the same as string passed to the constructor except that it will
   *  have the complement first eg. "complement(join(...
   **/
  public String toStringJoinInner () {
    final LocationParseNode top = getParsedLocation ();

    if (top.getType () == LocationParseNode.JOIN &&
        top.getJoinChildren ().elementAt (0).getType () ==
        LocationParseNode.COMPLEMENT) {
      final StringBuffer return_buffer = new StringBuffer ("complement(join(");

      final LocationParseNodeVector children = top.getJoinChildren ();

      // reverse the nodes because in a join(complement( location the
      // exons a in reverse order
      for (int i = children.size () - 1 ; i >= 0 ; --i) {

        final LocationParseNode complement_child =
          children.elementAt (i).getComplementChild ();

        return_buffer.append (complement_child.toString ());

        if (i != 0) {
          return_buffer.append (',');
        }
      }

      return_buffer.append ("))");

      return return_buffer.toString ();
    } else {
      // does not contain a join and a complement, so there is no need to swap
      // them
      return toString ();
    }
  }

  /**
   *  This method returns true iff this Location is a complement.
   **/
  public boolean isComplement () {
    if (getParsedLocation ().getType () == LocationParseNode.COMPLEMENT) {
      // COMPLEMENT(RANGE)
      return true;
    } else {
      if (getParsedLocation ().getType () == LocationParseNode.JOIN ||
          getParsedLocation ().getType () == LocationParseNode.ORDER) {
        // a join/order must have at least one child node
        if (getParsedLocation ().getChildren ().elementAt (0).getType () ==
            LocationParseNode.COMPLEMENT) {
          // JOIN(COMPLEMENT(RANGE)...) or ORDER(COMPLEMENT(RANGE)...)
          return true;
        } else {
          // JOIN(RANGE,...) or ORDER(RANGE,...)
          return false;
        }
      } else {
        // the head node is not a JOIN, COMPLEMENT or ORDER so it must be a
        // RANGE
        return false;
      }
    }
  }

  /**
   *  Return the reference of a new copy of this Location.
   **/
  public Location copy () {
    return new Location (getParsedLocation ().copy ());
  }

  /**
   *  Return a Range that spans the whole Location.
   **/
  public Range getTotalRange () {
    try {
      return new Range (getFirstBase (), getLastBase ());
    } catch (OutOfRangeException e) {
      throw new Error ("internal error - unexpected exception: " + e);
    }
  }

  /**
   *  Return the first Range of this Location, that is the range with the
   *  lowest start base.  Every location must have at least one Range.  For
   *  example if the location string was: <p>
   *  "join(complement(30..40),complement(10..20))"<p>
   *  then the returned Range will be 10..20
   **/
  public Range getFirstRange () {
    return getFirstOrLastRange (true);
  }

  /**
   *  Return the last Range of this Location, that is the range with the
   *  highest start base.  Every location must have at least one Range.  For
   *  example if the location string was: <p>
   *  "join(complement(30..40),complement(10..20))"<p>
   *  then the returned Range will be 30..40
   **/
  public Range getLastRange () {
    return getFirstOrLastRange (false);
  }

  /**
   *  Return the first (lowest) Range of this location of the argument is true
   *  or the last (highest) if the argument is false.
   *  @see #getFirstRange ()
   *  @see #getLastRange ()
   **/
  private Range getFirstOrLastRange (boolean get_first) {
    LocationParseNode test_node = null;
    if (getParsedLocation ().getType () == LocationParseNode.JOIN ||
        getParsedLocation ().getType () == LocationParseNode.ORDER) {

      final LocationParseNodeVector children =
        getParsedLocation ().getChildren ();

      if (isComplement () && get_first || !isComplement () && !get_first) {
        for (int i = children.size () - 1 ; i >= 0 ; --i) {
          final LocationParseNode this_child = children.elementAt (i);

          if (this_child.getType () == LocationParseNode.RANGE) {
            test_node = this_child;
            break;
          }

          if (this_child.getType () == LocationParseNode.COMPLEMENT &&
              this_child.getComplementChild ().getType () ==
              LocationParseNode.RANGE) {
            test_node = this_child.getComplementChild ();
            break;
          }
        }
      } else {
        for (int i = 0 ; i < children.size () ; ++i) {
          final LocationParseNode this_child = children.elementAt (i);

          if (this_child.getType () == LocationParseNode.RANGE) {
            test_node = this_child;
            break;
          }

          if (this_child.getType () == LocationParseNode.COMPLEMENT &&
              this_child.getComplementChild ().getType () ==
              LocationParseNode.RANGE) {
            test_node = this_child.getComplementChild ();
            break;
          }
        }
      }
    } else {
      if (getParsedLocation ().getType () == LocationParseNode.RANGE) {
        test_node = getParsedLocation ();
      }

      if (getParsedLocation ().getType () == LocationParseNode.COMPLEMENT) {
        test_node = getParsedLocation ().getComplementChild ();
      }
    }

    if (test_node == null) {
      // couldn't find a RANGE or COMPLEMENT
      return null;
    }

    return test_node.getRange ();
  }

  /**
   *  Return the first base that this Location covers.
   **/
  public int getFirstBase () {
    final Range first = getFirstRange ();

    if (first == null) {
      return -1;
    } else {
      return first.getStart ();
    }
  }

  /**
   *  Return the last base that this Location covers.
   **/
  public int getLastBase () {
    final Range last = getLastRange ();

    if (last == null) {
      return -1;
    } else {
      return last.getEnd ();
    }
  }

  /**
   *  Return a complemented copy of this Location.  If this Location is
   *  already a complement then a non complemented copy is returned.
   **/
  public Location getComplement () {
    return new Location (getParsedLocation ().getNodeComplement ());
  }

  /**
   *  Returns a parse tree for this Location.  Changes to the parse tree will
   *  change the original Location.
   *  @exception LocationParseException Thrown if a parse error occurs.
   **/
  public LocationParseNode getParsedLocation () {
    return parse_tree;
  }

  /**
   *  Create a new Location object from the given LocationParseNode tree.
   **/
  private Location (LocationParseNode parse_tree) {
    this.parse_tree = parse_tree;
  }

  /**
   *  Returns a parse tree for this Location.  Changes to the parse tree will
   *  not effect the original Location.
   *  @exception LocationParseException Thrown if a parse error occurs.
   */
  private LocationParseNode getParseTree (String location_string)
      throws LocationParseException {
    //    System.out.println ("in getParseTree () - location: " + location_string);

    final LocationLexer lexer = new LocationLexer (location_string);

    final TokenEnumeration enum = lexer.getTokens ();

    final LocationParseNode return_node = parseLocation (enum);

    if (enum.peekElement () != null) {
      throw new LocationParseException ("garbage at the end of the " +
                                        "location string",
                                        enum);
    }

    return return_node;
  }

  /* private functions: */

  /**
   *  Parse and return a location in the form of a LocationParseNode.  A
   *  location is of the form: complement(...), join(...), order(...) or
   *  <range> (where <range> is described in parseRange ()).  If the next
   *  tokens in the TokenEnumeration are not a location then an exception is
   *  thrown.  */
  private static LocationParseNode parseLocation (TokenEnumeration enum)
      throws LocationParseException {

    final LocationParseNode parsed_range = parseRange (enum);

    if (parsed_range != null) {
      return parsed_range;
    }

    // failed to parse a range - try parsing a functional (join, etc.)

    final LocationParseNode parsed_functional = parseFunctional (enum);

    if (parsed_functional != null) {
      return parsed_functional;
    }

    final LocationParseNode entry_range = parseEntryRange (enum);

    if (entry_range != null) {
      return entry_range;
    } else {
      throw new LocationParseException ("expected a range or a functional",
                                        enum);
    }
  }

  /**
   *  Attempt to parse and return a Range with an entry label in front (like
   *  J00194:(100..202) or J00193:hladr).  If the next tokens do not look like
   *  an entry range then it immediately returns null, otherwise an attempt is
   *  made to parse the next tokens as an entry range.  If the parse fails
   *  part way through then a LocationParseException is thrown.
   **/
  private static LocationParseNode parseEntryRange (TokenEnumeration enum)
      throws LocationParseException {
    if (! (enum.peekElement () instanceof String)) {
      return null;
    }

    final String entry_string = (String) enum.nextElement ();

    if (!enum.eatToken (':')) {
      throw new LocationParseException ("parse error after reading \"" +
                                        entry_string + "\"", enum);
    }

    final LocationParseNode range = parseRange (enum);

    return new LocationParseNode (entry_string, range);
  }

  /**
   *  Attempt to parse and return a Range.  If the next tokens do not look
   *  like a range then it immediately returns null, otherwise an attempt is
   *  made to read a range.  If the parse fails part way through what looks
   *  like a range then a LocationParseException is thrown.
   **/
  private static LocationParseNode parseRange (TokenEnumeration enum)
      throws LocationParseException {

    boolean is_lower_unbounded_range = false;
    boolean is_upper_unbounded_range = false;


    // returns Integer, UpperInteger, LowerInteger, Range or null
    Object start_object = parseRangeBound (enum);

    if (start_object == null) {
      return null;
    }

    final Object middle_token = enum.peekElement ();

    if (!enum.eatToken ("..") && !enum.eatToken ('^')) {
      return new LocationParseNode (FuzzyRange.makeRange (start_object));
    }

    if (enum.peekElement () == null) {
      throw new LocationParseException ("location ends in the middle of " +
                                        "a range", enum);
    }

    if (start_object instanceof UpperInteger) {
      throw new LocationParseException ("range cannot start with: " +
                                        start_object, enum);
    }

    Object end_object = parseRangeBound (enum);

    if (end_object == null) {
      throw new LocationParseException ("unexpected characters in location",
                                        enum);
    }

    if (end_object instanceof LowerInteger) {
      throw new LocationParseException ("a range cannot end with: " +
                                        end_object, enum);
    }

    if (middle_token.equals ("^")) {
      if (start_object instanceof Integer && end_object instanceof Integer) {
        try {
          final Range new_range =
            new BetweenRange (((Integer) start_object).intValue (),
                              ((Integer) end_object).intValue ());
          return new LocationParseNode (new_range);
        } catch (OutOfRangeException e) {
          throw new LocationParseException ("a range must start before it " +
                                            "ends", enum);
        }
      } else {
        throw new LocationParseException ("a range that contains a " +
                                          "'^' must start and end with a " +
                                          "plain number", enum);
      }
    }

    try {
      return new LocationParseNode (FuzzyRange.makeRange (start_object,
                                                          end_object));
    } catch (OutOfRangeException e) {
      throw new LocationParseException ("a range must start before it ends",
                                        enum);
    }
  }

  /**
   *  Parse the start or end of a range.
   **/
  private static Object parseRangeBound (TokenEnumeration enum) throws
      LocationParseException {
    if (enum.peekElement () instanceof Integer) {
      return enum.nextElement ();
    }

    if (enum.peekElement () instanceof LowerInteger) {
      return enum.nextElement ();
    }

    if (enum.peekElement () instanceof UpperInteger) {
      return enum.nextElement ();
    }

    // try to parse a range of the form (100.200)
    if (enum.eatToken ('(')) {
      if (enum.peekElement () instanceof Integer) {
        final Integer start = (Integer) enum.nextElement ();

        if (enum.eatToken ('.')) {
          if (enum.peekElement () instanceof Integer) {
            final Integer end = (Integer) enum.nextElement ();

            int start_value = start.intValue ();
            int end_value = end.intValue ();
            
            // make sure the ranges are the correct way around
            if (start_value > end_value) {
              final int tmp = start_value;
              start_value = end_value;
              end_value = tmp;
            }

            final Object return_object;

            if (start_value == end_value) {
              return_object = new Integer (start_value);
            } else {
              try {
                return_object = new Range (start_value, end_value);
              } catch (OutOfRangeException e) {
                throw new Error ("internal error - unexpected exception: " +e);
              }
            }

            if (enum.eatToken (')')) {
              return return_object;
            } else {
              final String message =
                "expected a closing parenthesis after reading: (" +
                start.intValue () + "." + end.intValue ();
              throw new LocationParseException (message, enum);
            }
          } else {
            throw new LocationParseException ("expected an integer", enum);
          }
        } else {
          throw new LocationParseException ("expected a '.'", enum);
        }
      } else {
        throw new LocationParseException ("expected an integer", enum);
      }
    }
    
    return null;
  }

  /**
   *  Attempt to parse and return a functional.  If the next tokens do not
   *  look like a functional then it immediately returns null, otherwise an
   *  attempt is made to read a functional.  If the parse fails part way
   *  through what looks like a functional then a LocationParseException is
   *  thrown.
   */
  private static LocationParseNode parseFunctional (TokenEnumeration enum)
      throws LocationParseException {
    Object peeked_element = enum.peekElement ();

    if (peeked_element instanceof String &&
        ("complement".equals((String) peeked_element) ||
         "join".equals((String) peeked_element) ||
         "order".equals((String) peeked_element))) {

      // remove the first token from the enum
      String functional_name = (String) enum.nextElement ();

      if (!enum.eatToken ('(')) {
        throw new LocationParseException ("expected (", enum);
      }

      if ("complement".equals (functional_name)) {
        // special case for complement - it should have only one argument

        LocationParseNode child = parseLocation (enum);

        if (!enum.eatToken (')')) {
          throw new LocationParseException ("expected )", enum);
        }

        // create a COMPLEMENT type node
        return new LocationParseNode (LocationParseNode.COMPLEMENT, child);
      } else {
        LocationParseNode child = parseLocation (enum);

        LocationParseNodeVector vector = new LocationParseNodeVector ();

        vector.addElement (child);

        while (enum.eatToken (',')) {
          LocationParseNode new_child = parseLocation (enum);

          vector.addElement (new_child);
        }

        if (!enum.eatToken (')')) {
          throw new LocationParseException ("expected )", enum);
        }

        if (vector.size () == 1) {
          final String message =
            "a " + functional_name + " operator must contain more " +
            "than one range";
          throw new LocationParseException (message, enum);
        }

        // create a JOIN or ORDER type node
        if ("join".equals (functional_name)) {
          return new LocationParseNode (LocationParseNode.JOIN, vector);
        } else {
          return new LocationParseNode (LocationParseNode.ORDER, vector);
        }
      }
    } else {
      // not a functional
      return null;
    }
  }

  /**
   *  The canonical parsed version of the location string that was passed to
   *  the constructor.
   **/
  private LocationParseNode parse_tree = null;
}
